#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup


setup(name='autoslot',
      version='2017.10.3',
      description='autoslot',
      author='Caleb Hattingh',
      author_email='caleb.hattingh@gmail.com',
      url='http://pythonomicon.com',
      py_modules=['autoslot'],
      python_requires='>=3.5',
     )
